@@ -9,8 +9,9 @@ use constant DEBUG => $ENV{SWAGGER2_DEBUG} || 0;
 
 my $SKIP_OP_RE = qr(By|From|For|In|Of|To|With);
 
-has url => '';
-has _validator => sub { Swagger2::SchemaValidator->new; };
+has url             => '';
+has _validator      => sub { Swagger2::SchemaValidator->new; };
+has _json_validator => sub { JSON::Validator->new; };
 
 sub dispatch_to_swagger {
   return undef unless $_[1]->{op} and $_[1]->{id} and ref $_[1]->{params} eq 'HASH';
@@ -100,8 +101,8 @@ sub register {
   $base_path =~ s!/$!!;
 
   for my $path (sort { length $a <=> length $b } keys %$paths) {
-    $paths->{$path}{'x-mojo-around-action'} ||= $swagger->api_spec->get('/x-mojo-around-action');
-    $paths->{$path}{'x-mojo-controller'}    ||= $swagger->api_spec->get('/x-mojo-controller');
+    my $around_action = $paths->{$path}{'x-mojo-around-action'} || $swagger->api_spec->get('/x-mojo-around-action');
+    my $controller    = $paths->{$path}{'x-mojo-controller'}    || $swagger->api_spec->get('/x-mojo-controller');
 
     for my $http_method (grep { !/^x-/ } keys %{$paths->{$path}}) {
       my $op_spec    = $paths->{$path}{$http_method};
@@ -114,8 +115,8 @@ sub register {
         "($type$name)";
       }/ge;
 
-      $op_spec->{'x-mojo-around-action'} ||= $paths->{$path}{'x-mojo-around-action'};
-      $op_spec->{'x-mojo-controller'}    ||= $paths->{$path}{'x-mojo-controller'};
+      $op_spec->{'x-mojo-around-action'} = $around_action if !$op_spec->{'x-mojo-around-action'} and $around_action;
+      $op_spec->{'x-mojo-controller'}    = $controller    if !$op_spec->{'x-mojo-controller'}    and $controller;
       $app->plugins->emit(
         swagger_route_added => $r->$http_method($route_path => $self->_generate_request_handler($op_spec)));
       warn "[Swagger2] Add route $http_method $base_path$route_path\n" if DEBUG;
@@ -125,7 +126,7 @@ sub register {
   if (my $spec_path = $config->{spec_path} // '/') {
     my $title = $swagger->api_spec->get('/info/title');
     $title =~ s!\W!_!g;
-    $r->get($spec_path)->to(cb => sub { shift->render(json => $swagger->api_spec->data) })->name(lc $title);
+    $r->get($spec_path)->to(cb => sub { _render_spec(shift, $swagger) })->name(lc $title);
   }
   if ($config->{ensure_swagger_response}) {
     $self->_ensure_swagger_response($app, $config->{ensure_swagger_response}, $swagger);
@@ -216,6 +217,16 @@ sub _on_route_added {
   $r->name($route_name);
 }
 
+sub _render_spec {
+  my ($c, $swagger) = @_;
+  my $spec = $swagger->api_spec->data;
+
+  local $spec->{id};
+  delete $spec->{id};
+  local $spec->{host} = $c->req->url->to_abs->host_port;
+  $c->render(json => $spec);
+}
+
 sub _validate_input {
   my ($self, $c, $op_spec) = @_;
   my (%cache, %input, @errors);
@@ -268,12 +279,18 @@ sub _validate_input_value {
 
   return if !defined $value and !Swagger2::_is_true($p->{required});
 
-  my $schema = {properties => {$name => $p->{schema} || $p}, required => [$p->{required} ? ($name) : ()]};
+  my $schema = {properties => {$name => $p->{'x-json-schema'} || $p->{schema} || $p},
+    required => [$p->{required} ? ($name) : ()]};
   my $in = $p->{in};
 
   if ($in eq 'body') {
     warn "[Swagger2] Validate $in $name\n" if DEBUG;
-    return $self->_validator->validate_input({$name => $value}, $schema);
+    if ($p->{'x-json-schema'}) {
+      return $self->_json_validator->validate({$name => $value}, $schema);
+    }
+    else {
+      return $self->_validator->validate_input({$name => $value}, $schema);
+    }
   }
   elsif (defined $value) {
     warn "[Swagger2] Validate $in $name=$value\n" if DEBUG;
@@ -309,7 +326,14 @@ sub _validate_response {
       }
     }
 
-    push @errors, $self->_validator->validate($data, $blueprint->{schema}) if $blueprint->{schema};
+    if ($blueprint->{'x-json-schema'}) {
+      warn "[Swagger2] Validate using x-json-schema\n" if DEBUG;
+      push @errors, $self->_json_validator->validate($data, $blueprint->{'x-json-schema'});
+    }
+    elsif ($blueprint->{schema}) {
+      warn "[Swagger2] Validate using schema\n" if DEBUG;
+      push @errors, $self->_validator->validate($data, $blueprint->{schema});
+    }
   }
   else {
     push @errors, $self->_validator->validate($data, {});
@@ -438,6 +462,8 @@ or look at L<Swagger2::Guides::Tutorial> for an introduction.
 
 =item * L<Swagger2::Guides::CustomPlaceholder> - Custom placeholder for your routes
 
+=item * L<Swagger2::Guides::JSONSchemaSupport> - Adding json-schema support - EXPERIMENTAL
+
 =item * L<Swagger spec|https://github.com/jhthorsen/swagger2/blob/master/t/blog/api.json>
 
 =item * L<Application|https://github.com/jhthorsen/swagger2/blob/master/t/blog/lib/Blog.pm>
